from anasymod.templates.generic_ip import TemplGenericIp
from anasymod.targets import FPGATarget
from anasymod.base_config import BaseConfig
from anasymod.enums import ConfigSections
from anasymod.config import EmuConfig

class TemplClkWiz(TemplGenericIp):
    def __init__(self, target: FPGATarget):

        ####################################################
        # Add module ports
        ####################################################


        ####################################################
        # Generate ip core config for clk wizard
        ####################################################
        props = {}
        # Add input clks
        props['CONFIG.PRIM_IN_FREQ'] = str(target.prj_cfg.board.clk_freq * 1e-6)
        if len(target.str_cfg.clk_i) == 2:
            props['CONFIG.PRIM_SOURCE'] = 'Differential_clock_capable_pin'
        elif len(target.str_cfg.clk_i) == 1:
            props['CONFIG.PRIM_SOURCE'] = 'Single_ended_clock_capable_pin'
        else:
            raise Exception("Wrong number of master clk pins is provided")

        # Add master output clk (emu_clk)
        props[f'CONFIG.CLKOUT1_USED'] = 'true'
        # commented out the line below because the "_PORT" config option is buggy
        props['CONFIG.CLKOUT1_REQUESTED_OUT_FREQ'] = (target.prj_cfg.cfg.emu_clk_freq * 1e-6)

        # Add debug clks
        for k, port in enumerate(target.str_cfg.clk_d):
            props[f'CONFIG.CLKOUT{k+2}_USED'] = 'true'
            # commented out the line below because the "_PORT" config option is buggy
            #props[f'CONFIG.CLKOUT{k+2}_PORT'] = port.name
            props[f'CONFIG.CLKOUT{k+2}_REQUESTED_OUT_FREQ'] = (target.prj_cfg.board.dbg_hub_clk_freq * 1e-6)

        # Add additional output clks
        #if self.target.str_cfg.cfg.clk_o_num:
        #    props['CONFIG.FEEDBACK_SOURCE'] = 'FDBK_AUTO'

        #for k, port in enumerate(self.target.str_cfg.clk_o):
        #    # commented out the line below because the "_PORT" config option is buggy
        #    #props[f'CONFIG.CLKOUT{k+self.target.str_cfg.clk_d_num+2}_PORT'] = port.name
        #    props[f'CONFIG.CLKOUT{k+self.target.str_cfg.clk_d_num+2}_USED'] = 'true'
        #    props[f'CONFIG.CLKOUT{k+self.target.str_cfg.clk_d_num+2}_REQUESTED_OUT_FREQ'] = (self.target.prj_cfg.cfg.emu_clk_freq * 1e-6)
        #    props[f'CONFIG.CLKOUT{k+self.target.str_cfg.clk_d_num+2}_DRIVES'] = 'BUFGCE'

        ####################################################
        # Prepare Template substitutions
        ####################################################

        props[f'CONFIG.NUM_OUT_CLKS'] = len(target.str_cfg.clk_m + target.str_cfg.clk_d) #ToDo: This shall not be hardwired!!!

        super().__init__(ip_name='clk_wiz', ip_dir=target.ip_dir, props=props)

def main():
    print(TemplClkWiz(target=FPGATarget(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()



from anasymod.templates.templ import JinjaTempl
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.sim_ctrl.ctrlifc_datatypes import DigitalCtrlInput, DigitalSignal
from anasymod.structures.structure_config import StructureConfig

class ModuleUARTSimCtrl(JinjaTempl):
    def __init__(self, scfg: StructureConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        custom_ctrl_ios = scfg.analog_ctrl_inputs + scfg.analog_ctrl_outputs + scfg.digital_ctrl_inputs + \
                          scfg.digital_ctrl_outputs

        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # Define module ios
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name="sim_ctrl_gen")
        module.add_inputs(ctrl_outputs)
        module.add_outputs(crtl_inputs + [scfg.dec_thr_ctrl] + [scfg.reset_ctrl])
        #ToDo: after revamp of clk in structure config, adding the master clk can be cleaned up as well
        module.add_input(DigitalCtrlInput(name=scfg.clk_m_ports[0].name, width=1, abspath=None))

        module.generate_header()
        #ToDo: currently clk will be treated as inputs to this wrapper, might be driven from PS at one point

        #####################################################
        # PC sim control section
        #####################################################

        # Custom control IOs for pc sim control module
        self.pc_sim_crtl_ifc = SVAPI()

        sim_ctrl_module = ModuleInst(api=self.pc_sim_crtl_ifc, name="sim_ctrl")
        sim_ctrl_module.add_inputs(ctrl_outputs, connections=ctrl_outputs)
        sim_ctrl_module.add_outputs(crtl_inputs, connections=crtl_inputs)

        sim_ctrl_module.generate_instantiation()

        # set number of clk cycles for initial reset
        self.rst_clkcycles = scfg.cfg.rst_clkcycles

        #####################################################
        # FPGA sim control section
        #####################################################

        # instantiation of register map module
        self.reg_map_inst = SVAPI()

        reg_map = ModuleInst(api=self.reg_map_inst, name="reg_map")

        reg_map.add_input(io_obj=DigitalSignal(name='clk', width=1, abspath=None), connection=DigitalSignal(name='emu_clk', width=1, abspath=None))

        rst = DigitalSignal(name='rst', width=1, abspath=None)
        reg_map.add_input(io_obj=rst, connection=scfg.reset_ctrl)

        i_addr = DigitalSignal(name='i_addr', width=8, abspath=None)
        reg_map.add_input(io_obj=i_addr, connection=i_addr)
        i_data = DigitalSignal(name='i_data', width=32, abspath=None)
        reg_map.add_input(io_obj=i_data, connection=i_data)
        o_addr = DigitalSignal(name='o_addr', width=8, abspath=None)
        reg_map.add_input(io_obj=o_addr, connection=o_addr)
        o_data = DigitalSignal(name='o_data', width=32, abspath=None)
        reg_map.add_output(io_obj=o_data, connection=o_data)

        reg_map.add_outputs(io_objs=crtl_inputs, connections=crtl_inputs)
        reg_map.add_inputs(io_objs=ctrl_outputs, connections=ctrl_outputs)

        reg_map.generate_instantiation()

        # instantiation of bd
        self.bd_inst = SVAPI()

        bd = ModuleInst(api=self.bd_inst, name="design_1_wrapper")

        i_addr_tri_o = DigitalSignal(name='i_addr_tri_o', width=8, abspath=None)
        i_data_tri_o = DigitalSignal(name='i_data_tri_o', width=8, abspath=None)
        o_addr_tri_o = DigitalSignal(name='o_addr_tri_o', width=8, abspath=None)
        o_data_tri_i = DigitalSignal(name='o_data_tri_i', width=8, abspath=None)
        ps_clk = DigitalSignal(name='ps_clk', width=8, abspath=None)
        reset_tri_o = DigitalSignal(name='reset_tri_o', width=8, abspath=None)


        bd.add_output(io_obj=i_addr_tri_o, connection=i_addr)
        bd.add_output(io_obj=i_data_tri_o, connection=i_data)
        bd.add_output(io_obj=o_addr_tri_o, connection=o_addr)
        bd.add_input(io_obj=o_data_tri_i, connection=o_data)
        bd.add_input(io_obj=ps_clk, connection=clk)
        # ToDo: Check direction of this clk
        bd.add_output(io_obj=reset_tri_o, connection=rst)

        bd.generate_instantiation()

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
    // reset sequence
    logic emu_rst_state = 1'b1;
    initial begin
        #((`DT_MSDSL)*{{subst.rst_clkcycles}}*1s);
        emu_rst_state = 1'b0;
    end

    // output assignment
    assign emu_rst = emu_rst_state;
    `ifndef DEC_THR_VAL_MSDSL
        `define DEC_THR_VAL_MSDSL 0
    `endif // `ifdef DEC_THR_VAL_MSDSL
    assign emu_dec_thr = `DEC_THR_VAL_MSDSL;

    {% if subst.pc_sim_crtl_ifc.text is not none %} 
    //module for custom vio handling
    //NOTE: sim_ctrl module must be written and added to the project manually!!!
{{subst.pc_sim_crtl_ifc.text}}
    {% endif %}
`else
    // Instantiation of register map
    {{subst.reg_map_inst.text}}

    // Instantiation of processing system
{{subst.bd_inst.text}}
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule
`default_nettype wire
'''

def main():
    from anasymod.config import EmuConfig
    #print(ModuleRegMapSimCtrl(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())
    print(ModuleUARTSimCtrl(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.ctrlifc_datatypes import DigitalCtrlInput, DigitalCtrlOutput, DigitalSignal, AnalogCtrlInput, AnalogCtrlOutput

class ModuleVIOSimCtrl(JinjaTempl):
    def __init__(self, scfg: StructureConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # Define module ios
        #####################################################
        self.module_ifc = SVAPI()

        module = ModuleInst(api=self.module_ifc, name="sim_ctrl_gen")
        module.add_inputs(ctrl_outputs)
        module.add_outputs(crtl_inputs + [scfg.dec_thr_ctrl] + [scfg.reset_ctrl])
        module.add_input(scfg.emu_clk)

        module.generate_header()

        #####################################################
        # PC sim control section
        #####################################################

        # Custom control IOs for pc sim control module
        self.pc_sim_crtl_ifc = SVAPI()

        sim_ctrl_module = ModuleInst(api=self.pc_sim_crtl_ifc, name="sim_ctrl")
        sim_ctrl_module.add_inputs(ctrl_outputs, connections=ctrl_outputs)
        sim_ctrl_module.add_outputs(crtl_inputs, connections=crtl_inputs)

        sim_ctrl_module.generate_instantiation()

        # set number of clk cycles for initial reset
        self.rst_clkcycles = scfg.cfg.rst_clkcycles

        #####################################################
        # FPGA sim control section - Instantiate vio wizard
        #####################################################

        self.vio_wiz_inst = SVAPI()
        vio_wiz = ModuleInst(api=self.vio_wiz_inst, name="vio_0")

        for k, input in enumerate(ctrl_outputs):
            if isinstance(input, AnalogCtrlOutput):
                width = '`LONG_WIDTH_REAL'
            elif isinstance(input, DigitalCtrlOutput):
                width = input.width
            else:
                raise Exception(f"Provided signal type:{type(input)} is not supported!")

            vio_i = DigitalSignal(name=f'probe_in{k}', width=width, abspath=None)
            vio_wiz.add_input(io_obj=vio_i, connection=input)

        for k, output in enumerate([scfg.reset_ctrl] + [scfg.dec_thr_ctrl] + crtl_inputs):
            if isinstance(output, AnalogCtrlInput):
                width = '`LONG_WIDTH_REAL'
            elif isinstance(output, DigitalCtrlInput):
                width = output.width
            else:
                raise Exception(f"Provided signal type:{type(output)} is not supported!")

            vio_o = DigitalSignal(name=f'probe_out{k}', width=width, abspath=None)
            vio_wiz.add_output(io_obj=vio_o, connection=output)

        vio_wiz.add_input(io_obj=DigitalSignal(name='clk', width=1, abspath=None), connection=scfg.emu_clk)
        vio_wiz.generate_instantiation()

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
    // reset sequence
    logic emu_rst_state = 1'b1;
    initial begin
        #((`DT_MSDSL)*{{subst.rst_clkcycles}}*1s);
        emu_rst_state = 1'b0;
    end

    // output assignment
    assign emu_rst = emu_rst_state;
    `ifndef DEC_THR_VAL_MSDSL
        `define DEC_THR_VAL_MSDSL 0
    `endif // `ifdef DEC_THR_VAL_MSDSL
    assign emu_dec_thr = `DEC_THR_VAL_MSDSL;

    {% if subst.pc_sim_crtl_ifc.text is not none %} 
    //module for custom vio handling
    //NOTE: sim_ctrl module must be written and added to the project manually!!!
{{subst.pc_sim_crtl_ifc.text}}
    {% endif %}
`else
	// VIO instantiation
{{subst.vio_wiz_inst.text}}
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule
`default_nettype wire
'''

def main():
    print(ModuleVIOSimCtrl(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.util import back2fwd
from anasymod.probe_config import ProbeConfig
from anasymod.targets import FPGATarget

class TemplEXECUTE_FPGA_SIM(JinjaTempl):
    def __init__(self, target: FPGATarget, start_time: float, stop_time: float, server_addr: str):
        super().__init__(trim_blocks=False, lstrip_blocks=False)
        pcfg = target.prj_cfg

        # read in probe signals from the probe config file
        self.probe_signals = ProbeConfig(probe_cfg_path=target.probe_cfg_path)

        # set server address
        self.server_addr = server_addr

        # set the paths to the BIT and LTX file
        self.bit_file = back2fwd(target.bitfile_path)
        self.ltx_file = back2fwd(target.ltxfile_path)

        # set the JTAG frequency.  sometimes it is useful to try a slower frequency than default if there
        # are problems with the debug hub clock
        self.jtag_freq = str(int(pcfg.cfg.jtag_freq))

        # set the "short" device name which is used to distinguish the FPGA part from other USB devices
        self.device_name = pcfg.board.short_part_name

        # set the path where the CSV file of results from the ILA should be written
        self.output = back2fwd(target.cfg.csv_path)
        self.ila_reset = target.prj_cfg.vivado_config.ila_reset
        #tbd remove vio_reset
        self.vio_reset = target.prj_cfg.vivado_config.vio_reset

        # set the window count to half the ILA depth.  this is required because the window depth will be "2"
        # unfortunately a window depth of "1" is not allowed in "BASIC" capture mode, which needed to allow
        # decimation of the sampling rate.
        self.window_count = str(pcfg.ila_depth/2)

        # calculate decimation ratio
        if stop_time is None:
            decimation_ratio_float = 2.0
        else:
            decimation_ratio_float = ((stop_time-start_time)/pcfg.cfg.dt)/(pcfg.ila_depth/2)

        # enforce a minimum decimation ratio of 2x since the window depth is 2
        decimation_ratio_float = max(decimation_ratio_float, 2.0)

        # we have to subtract 1 from decimation ratio due to the FPGA implementation of this feature
        # that is, a decimation ratio setting of "1" actually corresponds to a decimation factor of "2"; a setting
        # of "2" corresponds to a decimation ratio of "3", and so on.
        decimation_ratio_setting = int(round(decimation_ratio_float)) - 1

        # extract the time signal
        # note that time_signal is actually a list (should have exactly one value)
        if len(self.probe_signals.time_signal) == 0:
            raise Exception('Time signal not found -- check the probe config file.')
        elif len(self.probe_signals.time_signal) > 1:
            raise Exception('Multiple time signals not found  -- check the probe config file.')
        else:
            time_signal = self.probe_signals.time_signal[0]

        # extract properties from the time signal, which is a 3-tuple of strings
        time_name, time_width, time_exponent = time_signal
        time_width = int(time_width)
        time_exponent = int(time_exponent)

        # determine starting time as an integer value
        start_time_int = int(round(start_time*(2**(-time_exponent))))

        # export the decimation ratio, starting time, and time signal name to the template
        self.time_name = time_name
        self.decimation_ratio_setting = str(decimation_ratio_setting)
        self.start_time_int = f"{time_width}'u{start_time_int}"

    TEMPLATE_TEXT = '''
# Connect to hardware
open_hw
catch {disconnect_hw_server}
{% if subst.server_addr is none %}
connect_hw_server
{% else %}
connect_hw_server -url {{subst.server_addr}}
{% endif %}
set_property PARAM.FREQUENCY {{subst.jtag_freq}} [get_hw_targets]
open_hw_target

# Configure files to be programmed
set my_hw_device [get_hw_devices {{subst.device_name}}*]
current_hw_device $my_hw_device
refresh_hw_device $my_hw_device
set_property PROGRAM.FILE "{{subst.bit_file}}" $my_hw_device
set_property PROBES.FILE "{{subst.ltx_file}}" $my_hw_device
set_property FULL_PROBES.FILE "{{subst.ltx_file}}" $my_hw_device

# Program the device
program_hw_devices $my_hw_device
refresh_hw_device $my_hw_device

# VIO setup
set vio_0_i [get_hw_vios -of_objects $hw_device -filter {CELL_NAME=~"sim_ctrl_gen_i/vio_0_i"}]
set rst_hw_probe [get_hw_probes *rst* -of_objects $vio_0_i]

# Code related to non-interactive mode starts here 

# ILA setup
set my_hw_ila [get_hw_ilas]
display_hw_ila_data [get_hw_ila_data hw_ila_data_1 -of_objects $my_hw_ila]

# Trigger setup
# TODO: use starting time as trigger condition
startgroup
set_property CONTROL.CAPTURE_MODE BASIC $my_hw_ila
set_property CONTROL.TRIGGER_POSITION 0 $my_hw_ila
endgroup
set_property TRIGGER_COMPARE_VALUE gt{{subst.start_time_int}} [get_hw_probes {{subst.time_name}} -of_objects $my_hw_ila]

# Capture setup
# TODO: use variable for emu_dec_cmp_probe
# probably always be "2" because unfortunately "1" is not a valid option.
set_property CONTROL.DATA_DEPTH 2 $my_hw_ila
set_property CONTROL.WINDOW_COUNT {{subst.window_count}} $my_hw_ila
set_property CAPTURE_COMPARE_VALUE eq1'b1 [get_hw_probes emu_dec_cmp_probe -of_objects $my_hw_ila]

# VIO: decimation ratio
# TODO: use variable for emu_dec_thr
set_property OUTPUT_VALUE_RADIX UNSIGNED [get_hw_probes vio_gen_i/emu_dec_thr -of_objects $vio_0_i]
set_property OUTPUT_VALUE {{subst.decimation_ratio_setting}} [get_hw_probes vio_gen_i/emu_dec_thr -of_objects $vio_0_i]
commit_hw_vio [get_hw_probes vio_gen_i/emu_dec_thr -of_objects $vio_0_i]

# Radix setup: real numbers
{% for probename, _, _ in subst.probe_signals.analog_signals + subst.probe_signals.time_signal %}
catch {{'{'}}set_property DISPLAY_RADIX SIGNED [get_hw_probes {{probename}}]{{'}'}}
{% endfor %}

# Radix setup: unsigned digital values
# TODO: handle both signed and unsigned digital values
{% for probename, _, _ in subst.probe_signals.digital_signals + subst.probe_signals.reset_signal %}
catch {{'{'}}set_property DISPLAY_RADIX UNSIGNED [get_hw_probes {{probename}}]{{'}'}}
{% endfor %}

# Put design in reset
startgroup
set_property OUTPUT_VALUE 1 $rst_hw_probe
commit_hw_vio $rst_hw_probe
endgroup

# Arm the ILA trigger
run_hw_ila $my_hw_ila

# Take design out of reset
startgroup
set_property OUTPUT_VALUE 0 $rst_hw_probe
commit_hw_vio $rst_hw_probe
endgroup

# Get data from the ILA
wait_on_hw_ila $my_hw_ila
display_hw_ila_data [upload_hw_ila_data $my_hw_ila]

# Write ILA data to a file
write_hw_ila_data -csv_file -force "{{subst.output}}" hw_ila_data_1
'''

def main():
    print(TemplEXECUTE_FPGA_SIM(cfg=EmuConfig()).render())

if __name__ == "__main__":
    main()



import os

from anasymod.enums import ConfigSections
from anasymod.base_config import BaseConfig
from anasymod.config import EmuConfig
from anasymod.structures.port_base import PortIN, PortOUT, Port
from anasymod.structures.signal_base import Signal
from anasymod.sim_ctrl.ctrlifc_datatypes import DigitalSignal, DigitalCtrlInput, DigitalCtrlOutput, AnalogSignal, AnalogCtrlInput, AnalogCtrlOutput, ProbeSignal

#ToDo: wrap vios into classes to better cope with parameters such as width, name, abs_path, portobj, sigobj, ...

class StructureConfig():
    """
    In this configuration, all the toplevel information about the generated toplevel is included.
    It will be used for generation of the target specific top-level, as well as attached wrappers.

    There is also a specific interface to flow plugins that allows modification due to some needs
    from the plugin side, e.g. additional clks, resets, ios to the host application or resources on the FPGA board.
    """
    def __init__(self, prj_cfg: EmuConfig):
        # Internal variables
        self.i_addr_counter = 0
        self.o_addr_counter = 0

        # Path to ctrl_io file
        self._ctrl_iofile_path = os.path.join(prj_cfg.root, 'ctrl_io.config')
        # Path to clk file
        self._clk_file_path = os.path.join(prj_cfg.root, 'clk.config')
        # Path to clk file
        self._dt_req_file_path = os.path.join(prj_cfg.root, 'dt_req.config')
        # Path to probe file
        self._probe_file_path = os.path.join(prj_cfg.root, 'probe.config')


        self.cfg = Config(prj_cfg=prj_cfg)
        self.cfg.update_config()

        #########################################################
        # VIO interfaces
        #########################################################

        # Add DigitalCtrlInput for reset
        self.reset_ctrl = DigitalCtrlInput(abspath=None, name='emu_rst', width=1)
        self.reset_ctrl.i_addr = self._assign_i_addr()

        # Add DigitalCtrlInput for control signal 'emu_dec_thr' to manage decimation ration for capturing probe samples
        self.dec_thr_ctrl = DigitalCtrlInput(abspath=None, name='emu_dec_thr', width=int(prj_cfg.cfg.dec_bits))
        self.dec_thr_ctrl.i_addr = self._assign_i_addr()

        # CtrlIOs
        self.digital_ctrl_inputs = []
        self.digital_ctrl_outputs = []
        self.analog_ctrl_inputs = []
        self.analog_ctrl_outputs = []

        #Only for testing
        #self.digital_ctrl_inputs = [DigitalCtrlInput(name='hufflpu', width=31, abspath='test',init_value=41)]
        #self.digital_ctrl_inputs[0].i_addr = self._assign_i_addr()
        #self.digital_ctrl_outputs = [DigitalCtrlOutput(name='banii', width=1, abspath='testi')]
        #self.digital_ctrl_outputs[0].o_addr = self._assign_o_addr()
        #self.analog_ctrl_inputs = [AnalogCtrlInput(name='mimpfl', init_value=42.0, abspath='testii', range=500)]
        #self.analog_ctrl_inputs[0].i_addr = self._assign_i_addr()
        #self.analog_ctrl_outputs = [AnalogCtrlOutput(name='primpf', range=250, abspath=r'top.tb_i.v_out')]
        #self.analog_ctrl_outputs[0].o_addr = self._assign_o_addr()

        self._read_iofile()

        #########################################################
        # CLK manager interfaces
        #########################################################

        self.emu_clk = DigitalSignal(name='emu_clk', abspath=None, width=1)

        # add clk_in
        self.clk_i_num = len(prj_cfg.board.clk_pin)

        # clk_in names cannot be changed
        if self.clk_i_num == 2:
            self.clk_i = [DigitalSignal(abspath=None, width=1, name='clk_in1_p'), DigitalSignal(abspath=None, width=1, name='clk_in1_n')]
        elif self.clk_i_num == 1:
            self.clk_i = [DigitalSignal(abspath=None, width=1, name='clk_in1')]
        else:
            raise ValueError(
                f"Wrong number of pins for boards param 'clk_pin', expecting 1 or 2, provided:{self.clk_i_num}")

        # add master clk_outs, currently there is exactly one master clock, in case there is the need for multiple ones,
        # the num parameter needs to be exposed and naming needs to be configurable -> names might still be hardcoded at
        # some places!!!
        self.clk_m_num = 1
        self.clk_m = [DigitalSignal(abspath=None, width=1, name='emu_clk_2x')]

        # add debug clk_out
        self.clk_d_num = 1
        self.clk_d = []

        # currently there is exactly one debug clock, in case there is the need for multiple ones, name handling must be
        # implemented, there are some places in codebase where the debug clk name is still hard coded!!!
        for k in range(self.clk_d_num):
            self.clk_d += [DigitalSignal(abspath=None, width=1, name=f'dbg_hub_clk{k}')]

        #########################################################
        # EMU CLK generator interfaces
        #########################################################

        # add custom clk_out and associated clk_gate (currently all of them are derrived from this one master clk)
        # signals as tuples (clk_out, clk_gate)
        self.clk_o = []

        self._read_clkfile()

        #########################################################
        # Time manager interfaces
        #########################################################

        self.dt_reqs = []

        self._read_dt_reqfile()

        #########################################################
        # Probe interfaces
        #########################################################

        self.analog_probes = [] # use later
        self.time_probes = [] # use later
        self.reset_probes = [] # use later
        self.digital_probes = [] # use later

        #temporary
        self.probes = []
        """ : type: ProbeSignal"""

        self._read_probefile()

    def _assign_i_addr(self):
        """
        Function to assign an input address to an Input object.
        """
        curr_addr = self.i_addr_counter
        self.i_addr_counter +=1
        return curr_addr

    def _assign_o_addr(self):
        """
        Function to assign an input address to an Input object.
        """
        curr_addr = self.o_addr_counter
        self.o_addr_counter +=1
        return curr_addr

    def _read_iofile(self):
        """
        Read all lines from iofile and call parse function to populate ctrl_ios attributes.
        """
        if os.path.isfile(self._ctrl_iofile_path):
            with open(self._ctrl_iofile_path, "r") as f:
                ctrlios = f.readlines()
            self._parse_iofile(ctrlios=ctrlios)
        else:
            print(f"No ctrl_io file existing, no additional control IOs will be available for this simulation.")

    def _parse_iofile(self, ctrlios: list):
        """
        Read all lines from ctrl io file and store IO objects in CtrlIO container while adding access addresses to
        each IO object.
        :param ctrlios: Lines extracted to ctrl_io file
        """

        for k, line in enumerate(ctrlios):
            line = line.strip()

            if line.startswith('#'):
                # skip comments
                continue

            if line:
                try:
                    line = eval(line)
                    if isinstance(line, DigitalCtrlInput):
                        line.i_addr = self._assign_i_addr()
                        self.digital_ctrl_inputs.append(line)
                    elif isinstance(line, DigitalCtrlOutput):
                        line.o_addr = self._assign_o_addr()
                        self.digital_ctrl_outputs.append(line)
                    elif isinstance(line, AnalogCtrlInput):
                        line.i_addr = self._assign_i_addr()
                        self.analog_ctrl_inputs.append(line)
                    elif isinstance(line, AnalogCtrlOutput):
                        line.o_addr = self._assign_o_addr()
                        self.analog_ctrl_outputs.append(line)
                    else:
                        raise Exception(f"Line {k+1} of ctrl_io file: {self._ctrl_iofile_path} does not fit do a specified source or config type")
                except:
                    raise Exception(f"Line {k+1} of config file: {self._ctrl_iofile_path} could not be processed properely")

    def _read_clkfile(self):
        """
        Read all lines from clk.config file and call parse function to populate emu clk attribute.
        """
        if os.path.isfile(self._clk_file_path):
            with open(self._clk_file_path, "r") as f:
                clks = f.readlines()
            self._parse_clkfile(clks=clks)
        else:
            print(f"No clk.config file existing, no additional clks will be added.")

    def _parse_clkfile(self, clks: []):
        """
        Read all lines from ctrl io file and store IO objects in CtrlIO container while adding access addresses to
        each IO object.
        :param clks: Lines extracted from clk file
        """

        for k, item in enumerate(clks):
            item = item.strip()

            if item.startswith('#'):
                # skip comments
                continue

            if item:
                try:
                    item = eval(item)
                    if isinstance(item, str):
                        self.clk_o.append(item)
                    else:
                        raise Exception(f"Elements of line {k + 1} in clk file: {self._clk_file_path} don't consist of strings")
                except:
                    raise Exception(f"Line {k+1} of clk file: {self._clk_file_path} could not be processed properely")

    def _read_dt_reqfile(self):
        """
        Read all lines from dt_req.config file and call parse function to populate dt_req attribute.
        """
        if os.path.isfile(self._dt_req_file_path):
            with open(self._dt_req_file_path, "r") as f:
                dt_reqs = f.readlines()
            self._parse_dt_reqfile(dt_reqs=dt_reqs)
        else:
            print(f"No dt_req.config file existing, no additional dt requests will be added.")

    def _parse_dt_reqfile(self, dt_reqs: list):
        """
        Read all lines from dt_req file dt_req.config and store in dt_req attribute.
        :param clks: Lines extracted from dt_req file
        """

        for k, dt_req in enumerate(dt_reqs):
            dt_req = dt_req.strip()

            if dt_req.startswith('#'):
                # skip comments
                continue

            if dt_req:
                try:
                    dt_req = eval(dt_req)
                    if isinstance(dt_req, str):
                        self.dt_reqs.append(dt_req)
                    else:
                        raise Exception(f"Tuple elements of line {k + 1} in clk file: {self._dt_req_file_path} don't consist of strings")
                except:
                    raise Exception(f"Line {k+1} of clk file: {self._dt_req_file_path} could not be processed properely")

    def _read_probefile(self):
        """
        Read all lines from probe.config file and call parse function to populate probe attribute.
        """
        if os.path.isfile(self._probe_file_path):
            with open(self._probe_file_path, "r") as f:
                probes = f.readlines()
            self._parse_probefile(probes=probes)
        else:
            print(f"No probe.config file existing, no additional probes will be available for this simulation.")

    def _parse_probefile(self, probes: list):
        """
        Read all lines from probe file probe.config and store in probe attribute.
        :param clks: Lines extracted from dt_req file
        """

        for k, probe in enumerate(probes):
            probe = probe.strip()

            if probe.startswith('#'):
                # skip comments
                continue

            if probe:
                try:
                    probe = eval(probe)
                    if (isinstance(probe, list) and len(probe) == 4):
                        self.probes.append(ProbeSignal(name=probe[0], abspath=probe[1], width=probe[2], exponent=probe[3]))
                    else:
                        raise Exception(f"Probe specified in line {k + 1} in probe file: {self._probe_file_path} has "
                                        f"wrong format, expected is: ['name', 'abspath', 'width','exponent']")
                except:
                    raise Exception(f"Line {k+1} of probe.config file: {self._probe_file_path} could not be processed properely")

        # ToDo: Once different probe types are supported, parser needs to differenciate properly
        #self.analog_signals = []
        #for name, width, exponent in zip(signals[0], signals[2], signals[1]):
        #    self.analog_signals.append((name, width, exponent))

        #self.time_signal = []
        #for name,width, exponent in zip(signals[3], signals[5], signals[4]):
        #    self.time_signal.append((name, width, exponent))

        #self.reset_signal = []
        #for name,width, exponent in zip(signals[6], [r"1"], [None]):
        #    self.reset_signal.append((name, width, exponent))

        #self.digital_signals = []
        #for name in signals[7]:
        #    self.digital_signals.append((name, r"1", None))

        #for name, width in zip(signals[8], signals[9]):
        #    self.digital_signals.append((name, width, None))

class Config(BaseConfig):
    """
    Container to store all config attributes.
    """

    def __init__(self, prj_cfg: EmuConfig):
        super().__init__(cfg_file=prj_cfg.cfg_file, section=ConfigSections.STRUCTURE)

        self.rst_clkcycles = 1

        #########################################################
        # CLK manager settings
        #########################################################

        # add gated clk_outs
        self.clk_o_num = 0
        self.clk_g_num = 0


import os

from anasymod.viewer.viewer import Viewer
from anasymod.util import call

class SimVisionViewer(Viewer):
    def view(self):
       # build command
        cmd = [self.cfg.simvision_config.simvision, '-wave', self.target.cfg.vcd_path]

        # add waveform file if it exists
        if self.cfg.simvision_config.svcf_config is not None:
            if os.path.isfile(self.cfg.simvision_config.svcf_config):
                cmd.extend(['-input', self.cfg.simvision_config.svcf_config])

        # run command
        call(cmd)


import os.path

from anasymod.generators.vivado import VivadoTCLGenerator
from anasymod.generators.codegen import CodeGenerator
from anasymod.util import back2fwd

from anasymod.templates.dbg_hub import TemplDbgHub
from anasymod.templates.ext_clk import TemplExtClk
from anasymod.templates.clk_wiz import TemplClkWiz
from anasymod.templates.vio_wiz import TemplVIO
from anasymod.templates.execute_FPGA_sim import TemplEXECUTE_FPGA_SIM
from anasymod.templates.launch_FPGA_sim import TemplLAUNCH_FPGA_SIM
from anasymod.templates.probe_extract import TemplPROBE_EXTRACT
from anasymod.templates.ila import TemplILA
from anasymod.targets import FPGATarget
from anasymod.enums import FPGASimCtrl
from anasymod.sim_ctrl.ctrlinfra import ControlInfrastructure


class VivadoEmulation(VivadoTCLGenerator):
    """
    Generate and execute Vivado TCL scripts to generate a bitstream, run an emulation of FPGA for non-interactive mode,
    or launch an FPGA emulation for interactive mode and pass the handle for interactive control.
    """

    def __init__(self, target: FPGATarget):
        super().__init__(target=target)

    def build(self):
        # create a new project
        self.create_project(project_name=self.target.prj_cfg.vivado_config.project_name,
                              project_directory=self.target.project_root,
                              full_part_name=self.target.prj_cfg.board.full_part_name,
                              force=True)

        # add all source files to the project (including header files)
        self.add_project_sources(content=self.target.content)

        # define the top module
        self.set_property('top', f"{{{self.target.cfg.top_module}}}", '[current_fileset]')

        # set define variables
        self.add_project_defines(content=self.target.content, fileset='[current_fileset]')

        # append user constraints
        for xdc_file in self.target.content.xdc_files:
            for file in xdc_file.files:
                self.writeln(f'read_xdc "{back2fwd(file)}"')

        # write constraints to file
        constrs = CodeGenerator()
        constrs.use_templ(TemplExtClk(target=self.target))

        # TODO: allow tracing for custom_top
        if not self.target.cfg.custom_top:
            # generate clock wizard IP core
            self.use_templ(TemplClkWiz(target=self.target))

            # Add IP cores necessary for control interface
            ip_core_templates = self.target.ctrl.add_ip_cores(scfg=self.target.str_cfg, ip_dir=self.target.ip_dir)
            for ip_core_template in ip_core_templates:
                self.use_templ(ip_core_template)

            # Add constraints for additional generated emu_clks
            constrs.writeln('create_generated_clock -name emu_clk -source [get_pins clk_gen_i/clk_wiz_0_i/clk_out1] -divide_by 2 [get_pins gen_emu_clks_i/buf_emu_clk/I]')
            for k in range(len(self.target.str_cfg.clk_o)):
                constrs.writeln(f'create_generated_clock -name clk_other_{k} -source [get_pins clk_gen_i/clk_wiz_0_i/clk_out1] -divide_by 4 [get_pins gen_emu_clks_i/gen_other[{k}].buf_i/I]')

            # Setup ILA for signal probing
            self.use_templ(TemplILA(target=self.target))
    
            # Setup Debug Hub
            constrs.use_templ(TemplDbgHub(target=self.target))

        # write master constraints to file and add to project
        master_constr_path = os.path.join(self.target.prj_cfg.build_root, 'constrs.xdc')
        constrs.write_to_file(master_constr_path)
        self.add_files([master_constr_path], fileset='constrs_1')

        # read user-provided IPs
        self.writeln('# Custom user-provided IP cores')
        for xci_file in self.target.content.xci_files:
            for file in xci_file.files:
                self.writeln(f'read_ip "{back2fwd(file)}"')

        # upgrade IPs as necessary
        self.writeln('upgrade_ip [get_ips]')

        # generate all IPs
        self.writeln('generate_target all [get_ips]')

        # launch the build and wait for it to finish
        self.writeln(f'launch_runs impl_1 -to_step write_bitstream -jobs {min(int(self.target.prj_cfg.vivado_config.num_cores), 8)}')
        self.writeln('wait_on_run impl_1')

        # re-generate the LTX file
        # without this step, the ILA probes are sometimes split into individual bits
        ltx_file_path = os.path.join(self.target.project_root, f'{self.target.prj_cfg.vivado_config.project_name}.runs',
                                     'impl_1',
                                     f"{self.target.cfg.top_module}.ltx")
        self.writeln('open_run impl_1')
        self.writeln(f'write_debug_probes -force {{{back2fwd(ltx_file_path)}}}')

        # run bitstream generation
        self.run(filename=r"bitstream.tcl")

    def run_FPGA(self, start_time: float, stop_time: float, server_addr: str):
        """
        Run the FPGA in non-interactive mode. This means FPGA will run for specified duration, all specified signals
        will be captured and dumped to a file.

        :param start_time: Point in time from which recording run data will start
        :param stop_time: Point in time where FPGA run will be stopped
        :param dt: Update rate of analog behavior calculation
        :param server_addr: Hardware server address for hw server launched by Vivado
        """

        self.use_templ(TemplEXECUTE_FPGA_SIM(target=self.target, start_time=start_time, stop_time=stop_time, server_addr=server_addr))
        self.run(filename=r"run_FPGA.tcl", interactive=False)

    def launch_FPGA(self, server_addr: str):
        """
        Run the FPGA in interactive mode. This means FPGA will be programmed and control interfaces prepared. After that
        interactive communication with FPGA is possible.

        :param server_addr: Hardware server address for hw server launched by Vivado
        """

        self.use_templ(TemplLAUNCH_FPGA_SIM(target=self.target, server_addr=server_addr))
        self.run(filename=r"launch_FPGA.tcl", interactive=True)

        # ToDo: Depending on the fpga_ctrl setting in the project, setup the control interface e.g. UART or VIO, also
        # ToDo: check that VIO only works for linux

        # Return the control interface handle
        # ToDo: include spawnu + Steven's control setup to start Vivado TCL shell and return the shell handle
        #return



from anasymod.templates.templ import JinjaTempl
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.sim_ctrl.ctrlifc_datatypes import DigitalSignal


class ModuleRegMapSimCtrl(JinjaTempl):
    def __init__(self, scfg):
        super().__init__(trim_blocks=True, lstrip_blocks=True)

        crtl_inputs = scfg.analog_ctrl_inputs + scfg.digital_ctrl_inputs
        ctrl_outputs = scfg.analog_ctrl_outputs + scfg.digital_ctrl_outputs

        #####################################################
        # Define module ios
        #####################################################
        self.module_ifc = SVAPI()
        module = ModuleInst(api=self.module_ifc, name="reg_map")

        # Add clock
        clk = DigitalSignal(name='clk', width=1, abspath=None)
        module.add_input(io_obj=clk)

        # Add reset
        rst = DigitalSignal(name='rst', width=1, abspath=None)
        module.add_input(io_obj=rst)

        # Add I/Os to CPU subsystem
        i_addr = DigitalSignal(name='i_addr', width=8, abspath=None)
        module.add_input(io_obj=i_addr)
        i_data = DigitalSignal(name='i_data', width=32, abspath=None)
        module.add_input(io_obj=i_data)
        o_addr = DigitalSignal(name='o_addr', width=8, abspath=None)
        module.add_input(io_obj=o_addr)
        o_data = DigitalSignal(name='o_data', width=32, abspath=None)
        module.add_output(io_obj=o_data)

        # Add I/Os to Design (probes and Conrol Parameters)
        module.add_outputs(io_objs=crtl_inputs, connections=crtl_inputs)
        module.add_inputs(io_objs=ctrl_outputs, connections=ctrl_outputs)

        module.generate_header()
        #ToDo: currently clk will be treated as inputs to this wrapper, might be driven from PS at one point

        #####################################################
        # Initialize Default Values for ControlInfrastructure Parameters
        #####################################################

        self.init_ctrlios = SVAPI()

        for parameter in crtl_inputs:
            default_signal = parameter
            default_signal.name = str(default_signal.name) + '_def'
            self.init_ctrlios.gen_signal(io_obj=default_signal)
            self.init_ctrlios.assign_to(io_obj=default_signal, exp=default_signal.init_value)

        #####################################################
        # Combo mux for Probes section
        #####################################################

        # instantiation of register map module
        self.probes_combomux_cases = SVAPI()

        self.probes_combomux_cases.indent(quantity=3)
        for probe in ctrl_outputs:
            self.probes_combomux_cases.writeln(f"'d{probe.o_addr}: o_data = {probe.name};")

        #####################################################
        # Combo mux for Probes section
        #####################################################

        # instantiation of register map module
        self.params_regmap = SVAPI()

        for param in crtl_inputs:
            self.params_regmap.writeln(f"always @(posedge clk) begin")
            self.params_regmap.writeln(f"    if (rst == 'b1) begin")
            self.params_regmap.writeln(f"        {param.name} <= {param.name}_def; // use VIO defaults")
            self.params_regmap.writeln(f"    end else if (i_addr == 'd{param.i_addr}) begin ")
            self.params_regmap.writeln(f"        {param.name} <= i_data;")
            self.params_regmap.writeln(f"    end else begin")
            self.params_regmap.writeln(f"        {param.name} <= {param.name};")
            self.params_regmap.writeln(f"    end")
            self.params_regmap.writeln(f"end")
            self.params_regmap.writeln(f"")

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none

{{subst.module_ifc.text}}

    // Initial values for parameters
    {{subst.init_ctrlios.text}}  
    wire o_data_def;
    assign o_data_def = 0;

	// combo mux for reading outputs from design
	always @* begin
		case (o_addr)
{{subst.probes_combomux_cases.text}}
		    default: o_data = o_data_def;
		endcase
	end

	// register map for writing to the inputs of the design
	{{subst.params_regmap.text}}

endmodule
`default_nettype wire
'''

def main():
    from anasymod.structures.structure_config import StructureConfig
    from anasymod.config import EmuConfig
    print(ModuleRegMapSimCtrl(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


from anasymod.templates.templ import JinjaTempl
from anasymod.config import EmuConfig
from anasymod.generators.gen_api import SVAPI, ModuleInst
from anasymod.structures.structure_config import StructureConfig
from anasymod.sim_ctrl.ctrlifc_datatypes import DigitalSignal

class ModuleTracePort(JinjaTempl):
    def __init__(self, scfg: StructureConfig):
        super().__init__(trim_blocks=True, lstrip_blocks=True)
        # ToDo: This needs cleanup once probe class is conveniently implemented


        #####################################################
        # Define module ios
        #####################################################

        self.module_ifc = SVAPI()
        module = ModuleInst(api=self.module_ifc, name="trace_port_gen")
        # Add port signals
        for signal in scfg.probes:
            inst_sig = DigitalSignal(name=signal.name, abspath=signal.abspath, width=signal.width)
            module.add_input(inst_sig, connection=inst_sig)

        # Add master clk
        module.add_input(scfg.emu_clk)

        module.generate_header()

        #####################################################
        # FPGA sim control section - Instantiate vio wizard
        #####################################################

        self.ila_wiz_inst = SVAPI()
        ila_wiz = ModuleInst(api=self.ila_wiz_inst, name="ila_0")

        for k, signal in enumerate(scfg.probes):
            ila_wiz.add_input(DigitalSignal(name=f'probe{k}', abspath=None, width=signal.width),
                              connection=DigitalSignal(name=signal.name, abspath=signal.abspath, width=signal.width))

        # Add master clk
        ila_wiz.add_input(DigitalSignal(name='clk', abspath=None, width=1), connection=scfg.emu_clk)

        ila_wiz.generate_instantiation()

    TEMPLATE_TEXT = '''
`timescale 1ns/1ps

`default_nettype none
{{subst.module_ifc.text}}

`ifdef SIMULATION_MODE_MSDSL
    //Will be added later
`else
	// VIO instantiation
{{subst.ila_wiz_inst.text}}
`endif // `ifdef SIMULATION_MODE_MSDSL

endmodule
`default_nettype wire
'''

def main():
    print(ModuleTracePort(scfg=StructureConfig(prj_cfg=EmuConfig(root='test', cfg_file=''))).render())

if __name__ == "__main__":
    main()


import serial
import io, os
import serial.tools.list_ports as ports
from anasymod.enums import CtrlOps
from anasymod.sim_ctrl.ctrlinfra import ControlInfrastructure
from anasymod.structures.module_uartsimctrl import ModuleUARTSimCtrl
from anasymod.structures.module_regmapsimctrl import ModuleRegMapSimCtrl
from anasymod.sources import VerilogSource, BDFile
from anasymod.files import get_from_anasymod
from anasymod.structures.structure_config import StructureConfig

class UARTControlInfrastructure(ControlInfrastructure):
    def __init__(self, prj_cfg):
        super().__init__(prj_cfg=prj_cfg)

        # Initialize internal variables
        self._simctrlregmap_path = os.path.join(prj_cfg.build_root, 'gen_ctrlregmap.sv')

        # Program Zynq PS for UART access
        self._program_zynq_ps()
        #HIER WEITER: #ToDo: add path to elf file and add structure for storing ctrl ifc dependent files, also including the BD; later test if .tcl script for creating BD would be beneficial/at least there should be a script for creating the .bd file for a new Vivado version, also add binary path to xsct interface
        self.vid_list = [1027]
        self.pid_list = [24592]
        self.port_list = []

    #User functions
    def init_control(self):
        """
        Initialize the control interface, this is usually done after the bitstream was programmed successfully on the FPGA.
        """
        if self.cfg.comport is None: # run auto search for finding the correct COM port
            # find all available COM ports
            comports = [port for port in ports.comports(include_links=True)]
            # check if any COM port is compliant to known vids and pids and if so store the device_id
            for port in comports:
                if ((port.vid in self.vid_list) and (port.pid in self.pid_list)):
                    self.port_list.append(port.device)

            for port in self.port_list:
                try:
                    self._init_handler(comport=port)
                except:
                    pass
            if self.cfg.comport is None:
                raise Exception(f"ERROR: No COM port could be opened to enable connection to FPGA, found ports were: {self.port_list}.")

        else:
            try:
                self._init_handler(comport=self.cfg.comport)
            except:
                raise Exception(f"ERROR: Provided COM port: {self.cfg.comport} could not ne opened for communication.")

    def write_parameter(self, addr, data):
        self._write(operation=CtrlOps.WRITE_PARAMETER, addr=addr, data=data)
        if self._read():
            raise Exception(f"ERROR: Couldn't properly write: {addr}={data} command to FPGA.")

    def read_parameter(self, addr):
        self._write(operation=CtrlOps.READ_PARAMETER, addr=addr)
        return self._read()

    #Utility functions
    def _init_handler(self, comport):
        self.ctrl_handler = serial.Serial(comport, int(self.cfg.baud_rate), timeout=self.cfg.timeout,
                                          parity=self.cfg.parity, stopbits=self.cfg.stopbits,
                                          bytesize=self.cfg.bytesize)
        self.cfg.comport = comport

    def _write(self, operation, addr, data=None):
        # check is space is in any of the give input strings
        if ' ' in [operation, addr, data]:
            raise Exception(f"Blanks in any of the provided argument strings;{operation}, {addr}, {data}; sent via control interface are not allowed!")

        self.ctrl_handler.write((' '.join([str(operation), str(addr), str(data) + '\r']).encode('utf-8')))
        self.ctrl_handler.flush()

    def _read(self, count=1):
        for idx in range(count):
            result = self.ctrl_handler.readline()

            if result not in ['', None]:
                return int(result.decode("utf-8").rstrip())
        raise Exception(f"ERROR: Couldn't read from FPGA after:{count} attempts.")

    def gen_ctrlwrapper(self, str_cfg: StructureConfig, content):
        """
        Generate RTL design for control infrastructure. This will generate the register map, add the block diagram
        including the zynq PS and add the firmware running on the zynq PS.
        """

        # Generate simulation control wrapper and add to target sources
        with (open(self._simctrlwrap_path, 'w')) as ctrl_file:
           ctrl_file.write(ModuleUARTSimCtrl(scfg=str_cfg).render())

        content['verilog_sources'] += [VerilogSource(files=self._simctrlwrap_path)]

    def gen_ctrl_infrastructure(self, str_cfg: StructureConfig, content):
        """
        Generate RTL design for FPGA specific control infrastructure, depending on the interface selected for communication.
        For UART_ZYNQ control a register map, ZYNQ CPU SS block diagram and the firmware running on the zynq PS
        need to be handled.

        """

        # Generate register map according to IO settings stored in structure config and add to target sources
        with (open(self._simctrlregmap_path, 'w')) as ctrl_file:
           ctrl_file.write(ModuleRegMapSimCtrl(scfg=str_cfg).render())

        content['verilog_sources'] += [VerilogSource(files=self._simctrlregmap_path)]

        # Add ZYNQ CPU subsystem to target sources for UART IO as a blockdiagram
        zynq_bd = BDFile(files=get_from_anasymod('verilog', 'zynq_uart.bd'))
        content['bd_files'].append(zynq_bd)

        #ToDo: Add firmware part here -> generate FW if needed and add it to target sources

    def _program_zynq_ps(self):
        """
        Program UART control application to Zynq PS to enable UART control interface.
        """
        pass
        #HIER WEITER

    def add_ip_cores(self, scfg, ip_dir):
        """
        Configures and adds IP cores that are necessary for selected IP cores. No IP core is configured and added.
        :return rendered template for configuring a vio IP core
        """
        return []

def main():
    ctrl = UARTControlInfrastructure(prj_cfg=EmuConfig(root='test', cfg_file=''))
    ctrl.write_parameter(addr=0, data=3)
    ctrl.write_parameter(addr=1, data=4)
    print(ctrl.read_parameter(addr=0))
    print(ctrl.read_parameter(addr=1))

if __name__ == "__main__":
    main()